/**
 * Copyright (c) 2020 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 * @author Sebastian Bittrich <sebastian.bittrich@rcsb.org>
 */

import { ParamDefinition as PD } from '../../mol-util/param-definition';
import { StructureRepresentationPresetProvider, PresetStructureRepresentations } from '../../mol-plugin-state/builder/structure/representation-preset';
import { MembraneOrientationProvider, MembraneOrientation } from './prop';
import { StateObject, StateObjectRef, StateObjectCell, StateTransformer, StateTransform, StateBuilder } from '../../mol-state';
import { Task } from '../../mol-task';
import { PluginBehavior } from '../../mol-plugin/behavior';
import { MembraneOrientationRepresentationProvider, MembraneOrientationParams, MembraneOrientationRepresentation } from './representation';
import { HydrophobicityColorThemeProvider } from '../../mol-theme/color/hydrophobicity';
import { PluginStateObject, PluginStateTransform } from '../../mol-plugin-state/objects';
import { PluginContext } from '../../mol-plugin/context';
import { DefaultQueryRuntimeTable } from '../../mol-script/runtime/query/compiler';
import { StructureSelectionQuery, StructureSelectionCategory } from '../../mol-plugin-state/helpers/structure-selection-query';
import { MolScriptBuilder as MS } from '../../mol-script/language/builder';
import { GenericRepresentationRef } from '../../mol-plugin-state/manager/structure/hierarchy-state';
import { Expression } from '../../mol-script/language/expression';
import { StateTransforms } from '../../mol-plugin-state/transforms';
import { Color } from '../../mol-util/color';
import { createStructureRepresentationParams } from '../../mol-plugin-state/helpers/structure-representation-params';
import { PluginUIContext } from '../../mol-plugin-ui/context';
import { StateObjectSelector } from "../../mol-state/object";
import { MEMBRANE_STORAGE_KEY } from '../../extensions/tmdet/algorithm';
import { Vec3 } from '../../mol-math/linear-algebra';


const Tag = MembraneOrientation.Tag;

export const TMDETMembraneOrientation = PluginBehavior.create<{ autoAttach: boolean }>({
    name: 'tmdet-membrane-orientation-prop',
    category: 'custom-props',
    display: {
        name: 'Membrane Orientation',
        description: 'Data calculated with TMDET algorithm.'
    },
    ctor: class extends PluginBehavior.Handler<{ autoAttach: boolean }> {
        private provider = MembraneOrientationProvider

        register(): void {
            DefaultQueryRuntimeTable.addCustomProp(this.provider.descriptor);

            this.ctx.customStructureProperties.register(this.provider, this.params.autoAttach);

            this.ctx.representation.structure.registry.add(MembraneOrientationRepresentationProvider);
            this.ctx.query.structure.registry.add(isTransmembrane);

            this.ctx.genericRepresentationControls.set(Tag.Representation, selection => {
                const refs: GenericRepresentationRef[] = [];
                selection.structures.forEach(structure => {
                    const memRepr = structure.genericRepresentations?.filter(r => r.cell.transform.transformer.id === MembraneOrientation3D.id)[0];
                    if (memRepr) refs.push(memRepr);
                });
                return [refs, 'Membrane Orientation'];
            });
            this.ctx.builders.structure.representation.registerPreset(MembraneOrientationPreset);
        }

        update(p: { autoAttach: boolean }) {
            let updated = this.params.autoAttach !== p.autoAttach;
            this.params.autoAttach = p.autoAttach;
            this.ctx.customStructureProperties.setDefaultAutoAttach(this.provider.descriptor.name, this.params.autoAttach);
            return updated;
        }

        unregister() {
            DefaultQueryRuntimeTable.removeCustomProp(this.provider.descriptor);

            this.ctx.customStructureProperties.unregister(this.provider.descriptor.name);

            this.ctx.representation.structure.registry.remove(MembraneOrientationRepresentationProvider);
            this.ctx.query.structure.registry.remove(isTransmembrane);

            this.ctx.genericRepresentationControls.delete(Tag.Representation);
            this.ctx.builders.structure.representation.unregisterPreset(MembraneOrientationPreset);
        }
    },
    params: () => ({
        autoAttach: PD.Boolean(false)
    })
});

//

export const isTransmembrane = StructureSelectionQuery('Residues Embedded in Membrane', MS.struct.modifier.union([
    MS.struct.modifier.wholeResidues([
        MS.struct.modifier.union([
            MS.struct.generator.atomGroups({
                'chain-test': MS.core.rel.eq([MS.ammp('objectPrimitive'), 'atomistic']),
                'atom-test': MembraneOrientation.symbols.isTransmembrane.symbol(),
            })
        ])
    ])
]), {
    description: 'Select residues that are embedded between the membrane layers.',
    category: StructureSelectionCategory.Residue,
    ensureCustomProperties: (ctx, structure) => {
        return MembraneOrientationProvider.attach(ctx, structure);
    }
});

// //////////////////////////// TMDET VIEWER FUNCTIONS

let membrane: MembraneOrientation;


export async function loadWithUNITMPMembraneRepresentation(plugin: PluginUIContext, params: any) {
    const regionDescriptors: any = await downloadRegionDescriptor(plugin, params);

    membrane = createMembraneOrientation(regionDescriptors);

    console.log('DEBUG-02', membrane.planePoint2);

    localStorage.setItem(MEMBRANE_STORAGE_KEY, JSON.stringify(membrane));

    // load structure
    await loadStructure(plugin, params, regionDescriptors);

    await createStructureRepresentation(plugin, regionDescriptors);

    //
    // reset the camera because the membranes render 1st and the structure might not be fully visible
    //
    requestAnimationFrame(() => plugin.canvas3d?.requestCameraReset());
}

function createMembraneOrientation(regionDescriptors: any): MembraneOrientation {
    const membraneNormal: Vec3 = Vec3.fromObj(
        regionDescriptors['membrane-normal']
    );
    membraneNormal[0] = membraneNormal[2];
    membraneNormal[2] = 0;

    const membraneOrientation: MembraneOrientation = {
        planePoint1: Vec3.fromArray(Vec3.zero(), membraneNormal, 0),
        planePoint2: Vec3.fromArray(Vec3.zero(), membraneNormal, 0),
        centroid: Vec3.fromArray(
            Vec3.zero(), [0, 0, 0], 0
        ),
        normalVector: membraneNormal,

        // (NOTE: the TMDET extension calculates and sets it during applying preset)
        radius: regionDescriptors['radius']
    };
    membraneOrientation.planePoint2[0] *= -1;
    return membraneOrientation;
}

async function createStructureRepresentation(plugin: PluginUIContext, regionDescriptors: any) {
    // get the first structure of the first model
    const structure: StateObjectRef<PluginStateObject.Molecule.Structure> = plugin.managers.structure.hierarchy.current.models[0].structures[0].cell;
    const components = await createStructureComponents(plugin, structure);

    await buildStructureRepresentation(plugin, components);

    regionDescriptors.chains.forEach((chain: any) => {

        for (let regionKey in chain.regions) {
            const regionUpdates = plugin.build();
            const region = chain.regions[regionKey];
            createRegionRepresentation(plugin, chain.chain_id, region, regionUpdates.to(structure));
            regionUpdates.commit();
        }

    });
}

async function createStructureComponents(plugin: PluginUIContext, structure: StateObjectCell<PluginStateObject.Molecule.Structure, StateTransform<StateTransformer<StateObject<any, StateObject.Type<any>>, StateObject<any, StateObject.Type<any>>, any>>>) {
    return {
        polymer: await plugin.builders.structure.tryCreateComponentStatic(structure, 'polymer'),
        ligand: await plugin.builders.structure.tryCreateComponentStatic(structure, 'ligand'),
        water: await plugin.builders.structure.tryCreateComponentStatic(structure, 'water'),
    };
}

function createRegionRepresentation(plugin: PluginUIContext, chain: string, region: any, update: StateBuilder.To<any, any>) {
    const regionLabel: string = `${chain} | ${region.name}`;
    const color: Color = Color.fromArray(region.color, 0);
    const query: Expression = getQuery(chain, region.auth_ids as number[]);

    // based on https://github.com/molstar/molstar/issues/209
    update
        .apply(StateTransforms.Model.StructureSelectionFromExpression, { label: regionLabel, expression: query })
        .apply(StateTransforms.Representation.StructureRepresentation3D, createStructureRepresentationParams(plugin, update.selector.data, {
            color: 'uniform',
            colorParams: { value: color }
        }));
}

type StructureComponentType = StateObjectSelector<
        PluginStateObject.Molecule.Structure,
        StateTransformer<StateObject<any, StateObject.Type<any>>,
        StateObject<any, StateObject.Type<any>>, any>
    > | undefined;

async function buildStructureRepresentation(plugin: PluginUIContext, components: { polymer: StructureComponentType; ligand: StructureComponentType; water: StructureComponentType }) {
    const builder = plugin.builders.structure.representation;
    const update = plugin.build();
    if (components.polymer)
        builder.buildRepresentation(update, components.polymer, { type: 'cartoon', typeParams: { alpha: 0.5 } }, { tag: 'polymer' });
    if (components.ligand)
        builder.buildRepresentation(update, components.ligand, { type: 'ball-and-stick' }, { tag: 'ligand' });
    if (components.water)
        builder.buildRepresentation(update, components.water, { type: 'ball-and-stick', typeParams: { alpha: 0.6 } }, { tag: 'water' });
    await update.commit();
}

async function loadStructure(ctx: PluginUIContext, params: any, regionDescriptors: any): Promise<void> {
    const builders = ctx.builders;
    const data = await builders.data.download({
        url: params.structureUrl,
        label: `UniTMP: ${regionDescriptors['pdb-id']}`,
        isBinary: false
    }); // , { state: { isGhost: true } });
    const trajectory = await builders.structure.parseTrajectory(data, 'mmcif');
    // create membrane representation
    await builders.structure.hierarchy.applyPreset(
        trajectory, 'default', { representationPreset: 'preset-membrane-orientation' as any });
}

function getQuery(chainId: string, auth_array: number[]): Expression {
    const query: Expression =
        MS.struct.generator.atomGroups({
            'residue-test': MS.core.set.has([MS.set( ...auth_array ), MS.ammp('auth_seq_id')]),
            'chain-test': MS.core.rel.eq([chainId, MS.ammp('label_asym_id')])
        });
    return query;
}

async function downloadRegionDescriptor(plugin: PluginUIContext, params: any): Promise<any> {
    // run a fetch task
    const downloadResult: string = await plugin.runTask(plugin.fetch({ url: params.regionDescriptorUrl })) as string;
    const regionDescriptors: any = JSON.parse(downloadResult);
    return regionDescriptors;
}

// //////////////////////////// END OF TMDET VIEWER SECTION

//

export { MembraneOrientation3D };

type MembraneOrientation3D = typeof MembraneOrientation3D
const MembraneOrientation3D = PluginStateTransform.BuiltIn({
    name: 'membrane-orientation-3d',
    display: {
        name: 'Membrane Orientation',
        description: 'Membrane Orientation planes and rims. Data calculated with TMDET algorithm.'
    },
    from: PluginStateObject.Molecule.Structure,
    to: PluginStateObject.Shape.Representation3D,
    params: (a) => {
        return {
            ...MembraneOrientationParams,
        };
    }
})({
    canAutoUpdate({ oldParams, newParams }) {
        return true;
    },
    apply({ a, params }, plugin: PluginContext) {
        return Task.create('Membrane Orientation', async ctx => {
            await MembraneOrientationProvider.attach({ runtime: ctx, assetManager: plugin.managers.asset }, a.data);
            const repr = MembraneOrientationRepresentation({ webgl: plugin.canvas3d?.webgl, ...plugin.representation.structure.themes }, () => MembraneOrientationParams);
            await repr.createOrUpdate(params, a.data).runInContext(ctx);
            return new PluginStateObject.Shape.Representation3D({ repr, sourceData: a.data }, { label: 'Membrane Orientation' });
        });
    },
    update({ a, b, newParams }, plugin: PluginContext) {
        return Task.create('Membrane Orientation', async ctx => {
            await MembraneOrientationProvider.attach({ runtime: ctx, assetManager: plugin.managers.asset }, a.data);
            const props = { ...b.data.repr.props, ...newParams };
            await b.data.repr.createOrUpdate(props, a.data).runInContext(ctx);
            b.data.sourceData = a.data;
            return StateTransformer.UpdateResult.Updated;
        });
    },
    isApplicable(a) {
        return MembraneOrientationProvider.isApplicable(a.data);
    }
});

export const MembraneOrientationPreset = StructureRepresentationPresetProvider({
    id: 'preset-membrane-orientation',
    display: {
        name: 'Membrane Orientation', group: 'Annotation',
        description: 'Shows orientation of membrane layers. Data calculated with TMDET algorithm.' // TODO add ' or obtained via RCSB PDB'
    },
    isApplicable(a) {
        return MembraneOrientationProvider.isApplicable(a.data);
    },
    params: () => StructureRepresentationPresetProvider.CommonParams,
    async apply(ref, params, plugin) {
        const structureCell = StateObjectRef.resolveAndCheck(plugin.state.data, ref);
        const structure  = structureCell?.obj?.data;
        if (!structureCell || !structure) return {};

        if (!MembraneOrientationProvider.get(structure).value) {
            await plugin.runTask(Task.create('Membrane Orientation', async runtime => {
                await MembraneOrientationProvider.attach({ runtime, assetManager: plugin.managers.asset }, structure);
            }));
        }

        const membraneOrientation = await tryCreateMembraneOrientation(plugin, structureCell);
        const colorTheme = HydrophobicityColorThemeProvider.name as any;
        const preset = await PresetStructureRepresentations.auto.apply(ref, { ...params, theme: { globalName: colorTheme, focus: { name: colorTheme } } }, plugin);

        return { components: preset.components, representations: { ...preset.representations, membraneOrientation } };
    }
});

export function tryCreateMembraneOrientation(plugin: PluginContext, structure: StateObjectRef<PluginStateObject.Molecule.Structure>, params?: StateTransformer.Params<MembraneOrientation3D>, initialState?: Partial<StateTransform.State>) {
    const state = plugin.state.data;
    const membraneOrientation = state.build().to(structure)
        .applyOrUpdateTagged('membrane-orientation-3d', MembraneOrientation3D, params, { state: initialState });
    return membraneOrientation.commit({ revertOnError: true });
}
